#include "quakedef.h"
float TraceLine (vec3_t start, vec3_t end, vec3_t impact, vec3_t normal);
int numShadowLights;
int numStaticShadowLights;
int numUsedShadowLights; //number of shadow lights acutally drawn this frame

shadowlight_t shadowlights[MAXSHADOWLIGHTS];
shadowlight_t *usedshadowlights[MAXUSEDSHADOWLIGHTS];
shadowlightactive_t activeshadowlights[MAXUSEDSHADOWLIGHTS];
shadowlight_t *currentshadowlight;

cvar_t   sh_radiusscale = {"sh_radiusscale","0.5"};//PENTA: brightness of lightmaps

int numVolumeCmds;
int numLightCmds;
int numVolumeVerts;

msurface_t *shadowchain; //linked list of polygons that are shadowed
byte *lightvis;
byte worldvis[MAX_MAP_LEAFS/8];

/*
=============
AllocShadowLight
=============
*/
shadowlight_t* AllocShadowLight(void) {

   if (numShadowLights >= MAXSHADOWLIGHTS) {
      return NULL;
   }

   numShadowLights++;
   return &shadowlights[numShadowLights-1];
}

/*
=============
R_ShadowFromEntity
=============
*/
void R_ShadowFromEntity(entity_t *ent) {

   shadowlight_t *l;

   l = AllocShadowLight();
   if (!l) return;

   VectorCopy(ent->origin,l->origin);
   l->radius = 350;
   l->color[0] = 1;
   l->color[1] = 1;
   l->color[2] = 1;
   l->style = 0;
   l->brightness = 1;
   l->style = 0;
   l->style = ent->style;

   if (ent->light_lev != 0) {
      l->radius = ent->light_lev;
   } else {
      l->radius = 350;
   }

   VectorCopy(ent->color,l->baseColor);

   if ((l->baseColor[0] == 0) && (l->baseColor[1] == 0) && (l->baseColor[2] == 0)) {
      l->baseColor[0] = 1;
      l->baseColor[1] = 1;
      l->baseColor[2] = 1;
   }
}

float frustumPlanes[6][4];

float SphereInFrustum( vec3_t o, float radius )
{
   int p;
   float d;

   for( p = 0; p < 6; p++ )
   {
      d = frustumPlanes[p][0] * o[0] + frustumPlanes[p][1] * o[1] + frustumPlanes[p][2] * o[2] + frustumPlanes[p][3];
      if( d <= -radius )
         return 0;
   }
   return d + radius;
}

/*
=============
R_ContributeFrame

Returns true if the light should be rendered.
=============
*/
qboolean R_ContributeFrame (shadowlight_t *light)
{
   mleaf_t *lightleaf;
   float dist;
   vec3_t   hit = {0, 0, 0};

   float b = d_lightstylevalue[light->style]/255.0;

   light->color[0] = light->baseColor[0] * b;
   light->color[1] = light->baseColor[1] * b;
   light->color[2] = light->baseColor[2] * b;

   //frustum scissor testing
   dist = SphereInFrustum(light->origin, light->radius);
   if (dist == 0)  {
      //whole sphere is out ouf frustum so cut it.
      return false;
   }
   
   //trace a line from the eye to the light source
   TraceLine (r_refdef.vieworg, light->origin, hit,0);

   if (VectorLength(hit) != 0) {
      //simplicity in itself, this will keep number of active lights down
      //light cannot see player, it has hit a world object
      //do not include it
      return false;
   }

   return true;
}

/*
=============
R_ShadowFromDlight
=============
*/
void R_ShadowFromDlight(dlight_t *light)
{
   shadowlight_t *l;

   l = AllocShadowLight();
   if (!l) return;

   VectorCopy(light->origin,l->origin);
   l->radius = light->radius;
   l->color[0] = light->color[0];
   l->color[1] = light->color[1];
   l->color[2] = light->color[2];
   l->baseColor[0] = light->color[0];
   l->baseColor[1] = light->color[1];
   l->baseColor[2] = light->color[2];
   l->style = 0;
   l->brightness = 1;
   l->style = light->style;
}

/*
=============
R_MarkDLights

Adds dynamic lights to the shadow light list
=============
*/
void R_MarkDlights (void)
{
   int      i;
   dlight_t   *l;

   l = cl_dlights;
   for (i=0 ; i<MAX_DLIGHTS ; i++, l++)
   {
      if (l->die < cl.time || !l->radius)
         continue;
      R_ShadowFromDlight(l);
   }
}

int cut_ent;

void R_InitShadowsForFrame(void)
{
   byte *vis;
   int i;
   shadowlightactive_t *l = NULL;

   numShadowLights = numStaticShadowLights;

   R_MarkDlights (); //add dynamic lights to the list

   numUsedShadowLights = 0;

   cut_ent = 0;
   memset (&worldvis, 0, MAX_MAP_LEAFS/8); //all invisible

   vis = Mod_LeafPVS (r_viewleaf, cl.worldmodel);
   Q_memcpy(&worldvis, vis, MAX_MAP_LEAFS/8);

   for (i=0; i<numShadowLights; i++) {
      currentshadowlight = &shadowlights[i];
      if (R_ContributeFrame(currentshadowlight)) {
         currentshadowlight->visible = true;
         if (numUsedShadowLights < MAXUSEDSHADOWLIGHTS) {
            usedshadowlights[numUsedShadowLights] = currentshadowlight;
            l = &activeshadowlights[numUsedShadowLights];
            VectorCopy(currentshadowlight->origin, l->origin);
            l->radius = currentshadowlight->radius;
            numUsedShadowLights++;
         } else {
            Con_Printf("R_InitShadowsForFrame: More than MAXUSEDSHADOWLIGHS lights for frame\n");
         }
      } else {
         currentshadowlight->visible = false;
      }
   }
}

int done = 0;

/*
================
R_CalcSvBsp

Called for every static ent during spawning of the client
================
*/
void R_CalcSvBsp(entity_t *ent) {
   
   int i;
   msurface_t *surf;
   msurface_t   *s;
   texture_t   *t;

   if (ent->model == NULL) {
      Con_Printf("null model");
      return;
   }

   if (!strcmp (ent->model->name, "progs/flame2.mdl")
      || !strcmp (ent->model->name, "progs/flame.mdl")
      || !strcmp (ent->model->name, "progs/s_light.spr")
      || !strcmp (ent->model->name, "progs/b_light.spr")
      || !strcmp (ent->model->name, "progs/w_light.spr"))
   {
      shadowchain = NULL;
      done++;
      
      //Create a light and make it static
      R_ShadowFromEntity(ent);
      numStaticShadowLights++;
      
      if (numShadowLights >= MAXSHADOWLIGHTS)  {
         Con_Printf("R_CalcSvBsp: More than MAXSHADOWLIGHTS lights");
         return;
      }

      currentshadowlight = &shadowlights[numShadowLights-1];

      //Hack: support quake light_* entities
      if (!strcmp (ent->model->name, "progs/flame2.mdl")) {
         currentshadowlight->baseColor[0] = 1;
         currentshadowlight->baseColor[1] = 0.9;
         currentshadowlight->baseColor[2] = 0.75;   
      } else if (!strcmp (ent->model->name, "progs/flame.mdl")) {
         currentshadowlight->baseColor[0] = 1;
         currentshadowlight->baseColor[1] = 0.9;
         currentshadowlight->baseColor[2] = 0.75;
      } else if (!strcmp (ent->model->name, "progs/s_light.spr")) {
         currentshadowlight->baseColor[0] = 1;
         currentshadowlight->baseColor[1] = 0.9;
         currentshadowlight->baseColor[2] = 0.75;
      }

      //Con_Printf("light done\n");
   } else {
      Con_Printf("thrown away");
   }
}

#define surfaceLightRadius 350.0

void LightFromFile(vec3_t orig) 
{
   int i;
   shadowlight_t *light;
   qboolean   tooClose;
   entity_t   fakeEnt;
   vec3_t dist;
   float length;

   //not to close to other lights then add it

   tooClose = false;
   for (i=0; i<numShadowLights; i++) 
   {
   
	  light = &shadowlights[i];
      VectorSubtract(orig,light->origin,dist);
      length = VectorLength(dist);

      if (length < (surfaceLightRadius*sh_radiusscale.value + light->radius*sh_radiusscale.value)) 
	  {
         tooClose = true;
      }
   }


   if (!tooClose) {
      memset(&fakeEnt,0,sizeof(entity_t));
      fakeEnt.light_lev = surfaceLightRadius;
      VectorCopy(orig,fakeEnt.origin);
      fakeEnt.model = Mod_ForName("progs/w_light.spr",true);
      R_CalcSvBsp(&fakeEnt);
      //Con_Printf("Added file light");
   }


}


void ParseVector (char *s, float *d)
{
   int      i;
   char   string[128];
   char   *v, *w;
      
   strncpy (string, s,sizeof(string));
   v = string;
   w = string;
   for (i=0 ; i<3 ; i++)
   {
      while (*v && *v != ' ')
         v++;
      *v = 0;
      d[i] = atof (w);
      w = v = v+1;
   }
}



char *ParseEnt (char *data, qboolean *isLight, vec3_t origin)
{
   qboolean   init;
   char      keyname[256];
   qboolean   foundworld = false;
   init = false;

// go through all the dictionary pairs
   while (1)
   {   
   // parse key
      data = COM_Parse (data);
      if (com_token[0] == '}')
         break;
      if (!data)
         Sys_Error ("ParseEnt: EOF without closing brace");

      strncpy (keyname, com_token,sizeof(keyname));

   // parse value   
      data = COM_Parse (data);
      if (!data)
         Sys_Error ("ED_ParseEntity: EOF without closing brace");

      if (com_token[0] == '}')
         Sys_Error ("ED_ParseEntity: closing brace without data");

      init = true;   


      if (!strcmp(keyname, "classname")) {
         if (!strcmp(com_token, "light")) {
            *isLight = true;
         } else {
            *isLight = false;
         }
      } else if (!strcmp(keyname, "origin"))  {
         ParseVector(com_token, origin);   
      } else if (!strcmp(keyname, "_noautolight")) {
         Con_Printf("Automatic light gen disabled\n");//XYW \n
         foundworld = true;
      } else {

         //just do nothing

      }
   }

   if (foundworld) return NULL;
   return data;
}

void LoadLightsFromFile (char *data)
{   
   qboolean   isLight;
   vec3_t      origin;

// parse ents
   while (1)
   {
// parse the opening brace   
      data = COM_Parse (data);
      if (!data)
         break;
      if (com_token[0] != '{')
         Sys_Error ("ED_LoadFromFile: found %s when expecting {",com_token);

      isLight = false;
      data = ParseEnt (data, &isLight, origin);
      if (isLight) {
         LightFromFile(origin);
         //Con_Printf("found light in file");
      }
   }
}

void R_AutomaticLightPos()
{
   LoadLightsFromFile(cl.worldmodel->entities);
}